/*----------------------------------------------------------------------
  File    : wsom.c
  Contents: Windows program to visualize self-organizing map training
  Author  : Christian Borgelt
  History : 07.08.2000 file created from file wnn.c
            10.08.2000 first version completed
            13.07.2001 scaling factor for mouse movement introduced
            21.12.2001 cube region added
            2014.10.24 changed from LGPL license to MIT license
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <float.h>
#include <math.h>
#include <time.h>
#include <windows.h>
#include "wsom.rh"
#include "random.h"
#ifndef SOM_DRAW
#define SOM_DRAW
#endif
#include "somd.h"

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#ifndef M_PI                    /* if not defined (no Posix source) */
#define M_PI      3.14159265358979323846  /* \pi */
#endif

#define PRGNAME  "wsom"         /* main window title */

/* --- abbreviations --- */
#define HINST             HINSTANCE
#define MKIRSC            MAKEINTRESOURCE
#define SDIMSG            SendDlgItemMessage
#define WINSTYLE          WS_OVERLAPPEDWINDOW
#define PATH_MAX          _MAX_PATH

/* --- edit box functions --- */
#define edt_set(h,i,s)    SetDlgItemText(h, i, s)
#define edt_get(h,i,s,n)  GetDlgItemText(h, i, s, n)
#define int_set(h,i,n)    SetDlgItemInt(h, i, n, 0)
#define int_get(h,i)      GetDlgItemInt(h, i, NULL, 0)

/* --- combo box functions --- */
#define cmb_add(h,i,s)    SDIMSG(h, i, CB_ADDSTRING,    0, (LPARAM)(s))
#define cmb_rem(h,i,k)    SDIMSG(h, i, CB_DELETESTRING, k, 0)
#define cmb_set(h,i,k)    SDIMSG(h, i, CB_SETCURSEL,    k, 0)
#define cmb_get(h,i)      SDIMSG(h, i, CB_GETCURSEL,    0, 0)

/* --- extensions --- */
#define INITXEXT     320        /* initial x-extension of window */
#define INITYEXT     320        /* initial y-extension of window */

/* --- movement modes --- */
#define ROTATE         1        /* change heading and pitch */
#define YAW            2        /* change heading */
#define PITCH          3        /* change pitch angle */
#define ROLL           4        /* change roll  angle */
#define ZOOM           5        /* change distance to the origin */

/* --- error codes --- */
#define E_NONE         0        /* no error */
#define E_NOMEM      (-1)       /* not enough memory */
#define E_FOPEN      (-2)       /* cannot open file */
#define E_FREAD      (-3)       /* read error on file */
#define E_FWRITE     (-4)       /* write error on file */
#define E_INIT       (-5)       /* initialization failed */
#define E_SAVE       (-6)       /* no self-organizing map to save */
#define E_UNKNOWN    (-7)       /* unkown error */

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
const char *errmsgs[] = {       /* error messages */
  /* E_NONE      0 */  "no error",
  /* E_NOMEM    -1 */  "not enough memory",
  /* E_FOPEN    -2 */  "cannot open file\n%s",
  /* E_FREAD    -3 */  "read error on file\n%s",
  /* E_FWRITE   -4 */  "write error on file\n%s",
  /* E_INIT     -5 */  "initialization failed",
  /* E_SAVE     -6 */  "no self-organizing map created yet",
  /* E_UNKNOWN  -7 */  "unknown error"
};

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
static HINST   hinst;           /* instance handle */
static HWND    hmain;           /* handle of main window */
static HDC     hdc_win;         /* device context of window */
static HDC     hdc_buf;         /* device context of drawing buffer */
static HBITMAP buffer = NULL;   /* buffer for drawing */
static char    text[_MAX_PATH +1024];

/* --- pens and brushes --- */
static HPEN    pen_black;       /* black : grid of map */
static HPEN    pen_grey;        /* grey  : shape of pattern region */
static HBRUSH  br_white;        /* white : background */

/* --- neural network information --- */
static struct {                 /* -- parameters -- */
  int    shape;                 /* shape of pattern region */
  int    width, height;         /* width and height of the map */
  double range;                 /* initial weight range */
  double lrate;                 /* initial learning rate */
  double lrdec;                 /* learning rate decay parameter */
  int    nbhood;                /* neighborhood distance function */
  double radius;                /* initial neighborhood radius */
  double radec;                 /* radius decay parameter */
  int    patcnt;                /* patcnt between redraws */
  double delay;                 /* delay (in seconds) */
} nnpar = { SOM_SQUARE, 10, 10, 0.5, 0.6, -0.1,
            SOM_GAUSSIAN, 2.5, -0.1, 1, 0.05 };

static SOM    *som    = NULL;   /* current reference vectors, */
static int    patcnt  = 0;      /* current epoch */
static int    running = 0;      /* whether training is running */
static PROJ3D *proj3d = NULL;   /* 3D to 2D projection */
static int    mode    = 0;      /* movement mode */
static int    mx = 0, my = 0;   /* old mouse coordinates */
static double xoff, yoff;       /* coordinate offsets */
static double xscl, yscl;       /* scaling factors */
static HDC    hdc_draw;         /* device context for drawing */

/*----------------------------------------------------------------------
  Error Functions
----------------------------------------------------------------------*/

static void error (int code, ...)
{                               /* --- print error message */
  va_list     args;             /* list of variable arguments */
  const char  *msg;             /* error message */

  if ((code > 0) || (code < E_UNKNOWN))
    code = E_UNKNOWN;           /* check error code */
  va_start(args, code);         /* get variable arguments */
  msg = errmsgs[-code];         /* and error message text */
  if (!msg) msg = errmsgs[-E_UNKNOWN];
  vsprintf(text, msg, args);    /* format error message */
  MessageBox(hmain, text, PRGNAME, MB_OK|MB_ICONSTOP);
  va_end(args);                 /* show alert dialog box */
}  /* error() */

/*----------------------------------------------------------------------
  Drawing Functions
----------------------------------------------------------------------*/

static void drawpt (double x, double y, int line)
{                               /* --- move to/draw a line to a point */
  int cx, cy;                   /* coordinates of current point */

  cx = (int)(xoff +x *xscl);    /* transform the coordinates */
  cy = (int)(yoff -y *((nnpar.shape >= SOM_SPHERE) ? xscl : yscl));
  if (line) LineTo  (hdc_draw, cx, cy);
  else      MoveToEx(hdc_draw, cx, cy, NULL);
}  /* drawpt() */

/*--------------------------------------------------------------------*/

static void redraw (HDC hdc)
{                               /* --- Actions > Redraw */
  static int owd = 0, oht = 0;  /* old window extensions */
  static int opc = 0;           /* old pattern counter */
  RECT       rc;                /* rectangle for drawing */
  int        wd, ht;            /* window extensions */
  int        x0, x1, xm;        /* horizontal coordinates */
  int        y0, y1;            /* vertical coordinate */

  GetClientRect(hmain, &rc);    /* get the window's client rectangle */
  wd = rc.right  -rc.left +1;   /* and compute its width and height */
  ht = rc.bottom -rc.top  +1;   /* in pixels (it may have changed) */
  hdc_draw = hdc_buf;           /* default: draw into the buffer */
  if ((wd != owd) || (ht != oht)/* if the window extensions changed, */
  ||  !buffer) {                /* a new buffer has to be allocated */
    if (buffer) DeleteObject(buffer);
    buffer = CreateCompatibleBitmap(hdc_win, wd, ht);
    if (buffer) SelectObject(hdc_buf, buffer);
    else hdc_draw = hdc_win;    /* if no buffer can be allocated, */
  }                             /* draw directly into the window */
  owd = wd; oht = ht;           /* note the (new) window dimensions */
  xoff = 0.5*wd; yoff = 0.5*ht;
  xscl = 0.476*(wd-1);          /* compute the coordinate offsets */
  yscl = 0.476*(ht-1);          /* and the scaling factors */

  /* --- draw background --- */
  rc.left = rc.top = 0; rc.right = wd-1; rc.bottom = ht-1;
  FillRect(hdc_draw, &rc, br_white);

  /* --- draw pattern region --- */
  SelectObject(hdc_draw, pen_grey);
  x0 = (int)(xoff -xscl); x1 = (int)(xoff +xscl);
  y0 = (int)(yoff +yscl); y1 = (int)(yoff -yscl);
  xm = (int)(xoff);             /* compute coordinates */
  switch (nnpar.shape) {        /* evaluate shape */
    case SOM_SQUARE:            /* square region */
      MoveToEx(hdc_draw, x0, y0, NULL);
      LineTo  (hdc_draw, x1, y0);
      LineTo  (hdc_draw, x1, y1);
      LineTo  (hdc_draw, x0, y1);
      LineTo  (hdc_draw, x0, y0);         break;
    case SOM_TRIANGLE:          /* triangular region */
      MoveToEx(hdc_draw, x0, y0, NULL);
      LineTo  (hdc_draw, x1, y0);
      LineTo  (hdc_draw, xm, y1);
      LineTo  (hdc_draw, x0, y0);         break;
    case SOM_CIRCLE:            /* circular region */
      Ellipse (hdc_draw, x0, y1, x1, y0); break;
  }

  /* --- draw self-organizing map --- */
  if (som) {                    /* if there is a self-org. map */
    SelectObject(hdc_draw, pen_black);
    som_draw(som, (nnpar.shape >= SOM_SPHERE) ? proj3d : NULL, drawpt);
  }                             /* draw the self-organizing map */

  /* --- copy buffer to screen --- */
  if (hdc_draw == hdc_buf) {    /* if map was drawn to the buffer */
    if (!hdc) hdc = hdc_win;    /* get device context of window */
    BitBlt(hdc, 0, 0, wd-1, ht-1, hdc_buf, 0, 0, SRCCOPY);
  }                             /* copy the buffer to the window */

  /* --- show pattern counter --- */
  if (opc != patcnt) {          /* if the pattern count has changed */
    sprintf(text, "%s %10ld", PRGNAME, patcnt);
    SetWindowText(hmain, text); /* set a new window title */
    opc = patcnt;               /* with the new pattern count */
  }                             /* and note the new pattern count */
}  /* redraw() */

/*--------------------------------------------------------------------*/

static void mouse (int x, int y)
{                               /* --- evaluate mouse movement */
  int    dx, dy;                /* pixels mouse was moved */
  double sf;                    /* scaling factor */

  dx = x -mx;                   /* get the number of pixels */
  dy = y -my;                   /* the mouse was moved */
  sf = 64/xscl;                 /* compute the scaling factor */
  switch (mode) {               /* according to the movement mode */
    case ROTATE: if ((dx == 0) && (dy == 0)) return;
                 p3d_restore(proj3d);
                 p3d_pitch(proj3d, dy*sf*(M_PI/180), P3D_BOTH);
                 p3d_yaw  (proj3d, dx*sf*(M_PI/180), P3D_BOTH); break;
    case YAW:    if (dx == 0) return;
                 p3d_restore(proj3d);
                 p3d_yaw  (proj3d, dx*sf*(M_PI/180), P3D_BOTH); break;
    case PITCH:  if (dy == 0) return;
                 p3d_restore(proj3d);
                 p3d_pitch(proj3d, dy*sf*(M_PI/180), P3D_BOTH); break;
    case ROLL:   if (dx == 0) return;
                 p3d_restore(proj3d);
                 p3d_roll (proj3d, dx*sf*(M_PI/180), P3D_BOTH); break;
    case ZOOM:   if (dy == 0) return;
                 p3d_restore(proj3d);
                 p3d_move (proj3d, 0, dy*0.02, 0);              break;
  }                             /* change the viewing parameters */
}  /* mouse() */

/*----------------------------------------------------------------------
  Self-Organizing Map Training Function
----------------------------------------------------------------------*/

static void init (void)
{                               /* --- initialize self-org. map */
  if (!som) {                   /* if there is no self-org. map */
    som = som_create(nnpar.width, nnpar.height);
    if (!som) { error(E_NOMEM); return; }
  }                             /* create a self-organizing map */
  som_init(som, drand, nnpar.range, nnpar.range,
           (nnpar.shape >= SOM_SPHERE) ? nnpar.range : 0);
  patcnt = 0;                   /* clear the pattern counter and */
}  /* init() */

/*--------------------------------------------------------------------*/

static void CALLBACK train (HWND hwnd, UINT msg, UINT id, DWORD systime)
{                               /* --- do one training step */
  int    i;                     /* loop variable */
  double x, y, z;               /* pattern to train with */
  double lrate;                 /* current learning rate */
  double radius;                /* current neighborhood radius */

  if (!running || !som)         /* check whether training is running */
    return;                     /* and whether there is a map */
  for (i = nnpar.patcnt; --i >= 0; ) {
    som_genpat(nnpar.shape, drand, &x, &y, &z);
    patcnt++;                   /* count the training pattern */
    lrate  = (nnpar.lrdec <= 0) /* compute the learning rate */
           ? nnpar.lrate  *pow(patcnt, nnpar.lrdec)
           : nnpar.lrate  *pow(nnpar.lrdec, patcnt);
    radius = (nnpar.radec <= 0) /* compute the neighborhood radius */
           ? nnpar.radius *pow(patcnt, nnpar.radec)
           : nnpar.lrate  *pow(nnpar.lrdec, patcnt);
    som_train(som, x, y, z, lrate, nnpar.nbhood, radius);
  }                             /* train the self-organizing map */
  redraw(NULL);                 /* redraw the window contents */
  SetTimer(hwnd, 1, (UINT)(nnpar.delay *1000), (TIMERPROC)train);
}  /* train() */                /* reinstall timer */

/*----------------------------------------------------------------------
  Dialog Functions
----------------------------------------------------------------------*/

static char* fselect (char *title, int load)
{                               /* --- select a file */
  static char fname[_MAX_PATH+1];    /* file name */
  static char path [_MAX_PATH+1];    /* and path */
  static char filter[] =        /* file selection filter */
    "self-organizing maps (*.som)\0*.som\0" /* self-org. maps */
    "all files (*.*)\0*.*\0";               /* all files */
  static OPENFILENAME ofn = {   /* file selector structure */
    sizeof(OPENFILENAME),       /* lStructSize */
    NULL,   NULL,               /* hwndOwner, hInstance */
    filter, NULL,               /* lpstrFilter, lpstrCustomFilter */
    0, 1,                       /* nMaxCustFilter, nFilterIndex */
    fname, _MAX_PATH,           /* lpstrFile, nMaxFile */
    NULL, _MAX_FNAME+_MAX_EXT,  /* lpstrFileTitle, nMaxFileTitle */
    path,                       /* lpstrInitialDir */
    "Select File...",           /* lpstrTitle */
    OFN_PATHMUSTEXIST, 0, 0,    /* Flags, nFileOffset, nFileExtension */
    "nn", 0L,                   /* lpstrDefExt, lCustData */
    NULL, NULL };               /* lpfnHook, lpTemplateName */

  ofn.hwndOwner  = hmain;       /* set window handle */
  ofn.lpstrTitle = title;       /* and title text */
  ofn.Flags = (load) ? OFN_FILEMUSTEXIST : OFN_PATHMUSTEXIST;
  if (GetOpenFileName(&ofn) == 0)
    return NULL;                /* open file selector box */
  return fname;                 /* return name of selected file */
}  /* fselect() */

/*--------------------------------------------------------------------*/

static int load (void)
{                               /* --- File > Load Weights */
  char *fname;                  /* name of file to load */
  FILE *file;                   /* input file */
  SOM  *tmp;                    /* temporary buffer for s.o. map */
  int  shape;                   /* id of shape of pattern region */

  LoadString(hinst, CS_LOAD, text, 80);
  fname = fselect(text, 1);     /* get the dialog box title */
  if (!fname) return -1;        /* and select a file */
  file  = fopen(fname, "r");    /* open input file */
  if (!file) { error(E_FOPEN, fname); return -1; }
  if ((fscanf(file, "%i", &shape) != 1)
  ||  (shape < SOM_SQUARE)      /* read the id of the shape */
  ||  (shape > SOM_CUBE)) {     /* of the pattern region */
    error(E_FREAD, fname); return -1; }
  tmp = som_load(file);         /* load the self-org. map */
  fclose(file);                 /* close the input file */
  if (!tmp) { error(E_FREAD, fname); return -1; }
  if (som) som_delete(som);     /* delete an existing self-org. map */
  som         = tmp;            /* and set the loaded one */
  nnpar.shape = shape;          /* set shape of the pattern region */
  patcnt      = 0;              /* clear the pattern counter and */
  return 0;                     /* return 'ok' */
}  /* load() */

/*--------------------------------------------------------------------*/

static void save (void)
{                               /* --- File > Save Weights... */
  char *fname;                  /* name of file to save to */
  FILE *file;                   /* output file */

  if (!som) {                   /* if there is no map, */
    error(E_SAVE); return; }    /* abort the function */
  LoadString(hinst, CS_SAVE, text, 80);
  fname = fselect(text, 0);     /* get the dialog box title */
  if (!fname) return;           /* and select a file */
  file = fopen(fname, "w");     /* open output file */
  if (!file) { error(E_FOPEN, fname); return; }
  fprintf(file, "%d ", nnpar.shape);
  som_save(som, file);          /* save id of shape of pattern region */
  if (fclose(file) != 0)        /* and the self-organizing map, */
    error(E_FWRITE);            /* then close the output file */
}  /* save() */

/*--------------------------------------------------------------------*/

static void mn_enable (int enable)
{                               /* --- enable/disable menu items */
  HMENU        menu;            /* handle of window menu */
  MENUITEMINFO mii;             /* for menu customization */
  UINT         flag;            /* enable/disable flag */

  menu = GetMenu(hmain);        /* get menu handle */
  flag = (enable) ? MF_ENABLED : MF_GRAYED;
  EnableMenuItem(menu, MI_LOAD, flag);    /* enable/disable context */
  EnableMenuItem(menu, MI_GRID, flag);    /* sensitive menu items */
  mii.cbSize     = sizeof(MENUITEMINFO);  /* set either the */
  mii.fMask      = MIIM_ID|MIIM_TYPE;     /* 'Start Training' or */
  mii.fType      = MFT_STRING;            /* 'Stop Training' item */
  mii.wID        =  (enable) ? MI_START : MI_STOP;
  LoadString(hinst, (enable) ? CS_START : CS_STOP, text, 80);
  mii.dwTypeData = text; mii.cch = (UINT)strlen(text);
  SetMenuItemInfo(menu, (enable) ? MI_STOP : MI_START, FALSE, &mii);
}  /* mn_enable() */

/*----------------------------------------------------------------------
  Dialog Box Procedures
----------------------------------------------------------------------*/

void dbl_set (HWND hwnd, int item, double num)
{                               /* --- set floating point number */
  sprintf(text, "%g", num);     /* format floating point number */
  edt_set(hwnd, item, text);    /* set formatted number */
}  /* dbl_set() */

/*--------------------------------------------------------------------*/

double dbl_get (HWND hwnd, int item)
{                               /* --- get floating point number */
  edt_get(hwnd, item, text, 64);    /* get input text and */
  return strtod(text, NULL);    /* convert it to a number */
}  /* dbl_get() */

/*--------------------------------------------------------------------*/

BOOL CALLBACK dp_about (HWND hwnd, UINT msg, WPARAM wp, LPARAM lp)
{                               /* --- 'about' dialog box procedure */
  switch (msg) {                /* evaluate message */
    case WM_COMMAND:            /* --- if command */
      if (wp != IDOK) break;    /* process only 'Ok' button */
    case WM_CLOSE:              /* --- if dialog closed */
      EndDialog(hwnd, 0);       /* terminate dialog */
      return 1;                 /* return 'ok' */
  }
  return 0;                     /* everything else not handled */
}  /* dp_about() */

/*--------------------------------------------------------------------*/

BOOL CALLBACK dp_grid (HWND hwnd, UINT msg, WPARAM wp, LPARAM lp)
{                               /* --- grid/shape dialog procedure */
  switch (msg) {                /* evaluate message */
    case WM_INITDIALOG:         /* dialog initialization */
      LoadString(hinst, CS_SQUARE,   text, 80);
      cmb_add(hwnd, DI_SHAPE, text);
      LoadString(hinst, CS_TRIANGLE, text, 80);
      cmb_add(hwnd, DI_SHAPE, text);
      LoadString(hinst, CS_CIRCLE,   text, 80);
      cmb_add(hwnd, DI_SHAPE, text);
      LoadString(hinst, CS_SPHERE,   text, 80);
      cmb_add(hwnd, DI_SHAPE, text);
      LoadString(hinst, CS_PARABOLA, text, 80);
      cmb_add(hwnd, DI_SHAPE, text);
      LoadString(hinst, CS_CUBIC,    text, 80);
      cmb_add(hwnd, DI_SHAPE, text);
      LoadString(hinst, CS_COSINE,   text, 80);
      cmb_add(hwnd, DI_SHAPE, text);
      LoadString(hinst, CS_CUBE,     text, 80);
      cmb_add(hwnd, DI_SHAPE, text);
      switch (nnpar.shape) {    /* shape of the pattern region */
        case SOM_TRIANGLE: cmb_set(hwnd, DI_SHAPE, 1); break;
        case SOM_CIRCLE  : cmb_set(hwnd, DI_SHAPE, 2); break;
        case SOM_SPHERE  : cmb_set(hwnd, DI_SHAPE, 3); break;
        case SOM_PARABOLA: cmb_set(hwnd, DI_SHAPE, 4); break;
        case SOM_CUBIC   : cmb_set(hwnd, DI_SHAPE, 5); break;
        case SOM_COSINE  : cmb_set(hwnd, DI_SHAPE, 6); break;
        case SOM_CUBE    : cmb_set(hwnd, DI_SHAPE, 7); break;
        default          : cmb_set(hwnd, DI_SHAPE, 0); break;
      }                         /* width and height of map */
      int_set(hwnd, DI_WIDTH,  nnpar.width);
      int_set(hwnd, DI_HEIGHT, nnpar.height);
      break;
    case WM_COMMAND:            /* command */
      switch (wp) {             /* evaluate command code */
        case IDOK:              /* if 'Ok' button pressed */
          switch ((int)cmb_get(hwnd, DI_SHAPE)) {
            case  1: nnpar.shape = SOM_TRIANGLE; break;
            case  2: nnpar.shape = SOM_CIRCLE;   break;
            case  3: nnpar.shape = SOM_SPHERE;   break;
            case  4: nnpar.shape = SOM_PARABOLA; break;
            case  5: nnpar.shape = SOM_CUBIC;    break;
            case  6: nnpar.shape = SOM_COSINE;   break;
            case  7: nnpar.shape = SOM_CUBE;     break;
            default: nnpar.shape = SOM_SQUARE;   break;
          }                     /* shape of the pattern region */
          nnpar.width  = int_get(hwnd, DI_WIDTH);
          if      (nnpar.width  < 1)   nnpar.width  = 1;
          else if (nnpar.width  > 100) nnpar.width  = 100;
                                /* width  of self-org. map (in units) */
          nnpar.height = int_get(hwnd, DI_HEIGHT);
          if      (nnpar.height < 1)   nnpar.height = 1;
          else if (nnpar.height > 100) nnpar.height = 100;
                                /* height of self-org. map (in units) */
          if (som && ((som_width(som)  != nnpar.width)
          ||          (som_height(som) != nnpar.height))) {
            som_delete(som);    /* delete the self-organizing map */
            som = NULL;         /* if it does not have the same size */
          }                     /* as the requested one */
          EndDialog(hwnd, 0);   /* get inputs from edit controls */
          return 1;             /* return 'ok' */
        case IDCANCEL:          /* if 'Cancel' button pressed, */
          EndDialog(hwnd, -1); return 1;   /* terminate dialog */
      } break;
    case WM_CLOSE:              /* dialog box has been closed */
      EndDialog(hwnd, -1); return 1;      /* terminate dialog */
  }
  return 0;                     /* everything else not handled */
}  /* dp_grid() */

/*--------------------------------------------------------------------*/

BOOL CALLBACK dp_params (HWND hwnd, UINT msg, WPARAM wp, LPARAM lp)
{                               /* --- parameters dialog procedure */
  switch (msg) {                /* evaluate message */
    case WM_INITDIALOG:         /* dialog initialization */
      dbl_set(hwnd, DI_RANGE,  nnpar.range);
      dbl_set(hwnd, DI_LRATE,  nnpar.lrate);
      dbl_set(hwnd, DI_LRDEC,  nnpar.lrdec);
      LoadString(hinst, CS_RECT, text, 80);
      cmb_add(hwnd, DI_NBHOOD, text);
      LoadString(hinst, CS_GAUSSIAN,  text, 80);
      cmb_add(hwnd, DI_NBHOOD, text);
      cmb_set(hwnd, DI_NBHOOD, (nnpar.nbhood == SOM_RECT) ? 0 : 1);
      dbl_set(hwnd, DI_RADIUS, nnpar.radius);
      dbl_set(hwnd, DI_RADEC,  nnpar.radec);
      int_set(hwnd, DI_PATCNT, nnpar.patcnt);
      dbl_set(hwnd, DI_DELAY,  nnpar.delay);
      break;                    /* set learning parameters */
    case WM_COMMAND:            /* command */
      switch (wp) {             /* evaluate command code */
        case IDOK:              /* if 'Ok' button pressed */
          nnpar.range = dbl_get(hwnd, DI_RANGE);
          if      (nnpar.range < 0)     nnpar.range  = 0;
          else if (nnpar.range > 100)   nnpar.range  = 100;
                                /* initial weight range */
          nnpar.lrate = dbl_get(hwnd, DI_LRATE);
          if      (nnpar.lrate < 1e-12) nnpar.lrate  = 1e-12;
          else if (nnpar.lrate > 10)    nnpar.lrate  = 10;
                                /* inital learning rate */
          nnpar.lrdec  = dbl_get(hwnd, DI_LRDEC);
          if      (nnpar.lrdec  > 1)    nnpar.lrdec  = 1;
                                /* learning rate decay parameter */
          nnpar.nbhood = (cmb_get(hwnd, DI_NBHOOD) == 0)
                       ? SOM_RECT : SOM_GAUSSIAN;
                                /* neighborhood type */
          nnpar.radius = dbl_get(hwnd, DI_RADIUS);
          if      (nnpar.radius < 0)    nnpar.radius = 0;
          else if (nnpar.radius > 100)  nnpar.radius = 100;
                                /* initial neighborhood radius */
          nnpar.radec  = dbl_get(hwnd, DI_RADEC);
          if      (nnpar.radec  > 1)    nnpar.radec  = 1;
                                /* radius decay parameter */
          nnpar.patcnt = int_get(hwnd, DI_PATCNT);
          if      (nnpar.patcnt < 1)    nnpar.patcnt = 1;
          else if (nnpar.patcnt > 1000) nnpar.patcnt = 1000;
                                /* patterns between redraws */
          nnpar.delay = dbl_get(hwnd, DI_DELAY);
          if      (nnpar.delay < 0.001) nnpar.delay  = 0.001;
          else if (nnpar.delay > 10)    nnpar.delay  = 10;
                                /* delay between redraws */
          EndDialog(hwnd, 0);   /* terminate dialog and */
          return 1;             /* return 'ok' */
        case IDCANCEL:          /* if 'Cancel' button pressed, */
          EndDialog(hwnd, -1); return 1;   /* terminate dialog */
      } break;
    case WM_CLOSE:              /* dialog box has been closed */
      EndDialog(hwnd, -1); return 1;      /* terminate dialog */
  }
  return 0;                     /* everything else not handled */
}  /* dp_params() */

/*----------------------------------------------------------------------
  Window Procedure
----------------------------------------------------------------------*/

LRESULT CALLBACK wp_main (HWND hwnd, UINT msg, WPARAM wp, LPARAM lp)
{                               /* --- main window procedure */
  static PAINTSTRUCT ps;        /* for BeginPaint call */
  HDC hdc;                      /* handle of device context */

  switch (msg) {                /* evaluate message */
    case WM_PAINT:              /* --- if to paint window */
      hdc = BeginPaint(hwnd, &ps);
      if (!hdc) return 0;       /* get device context for drawing */
      redraw(hdc);              /* redraw the window contents */
      EndPaint(hwnd, &ps);      /* and release the device context */
      break;

    case WM_COMMAND:            /* --- if menu item selected */
      switch (wp) {             /* evaluate item */
        case MI_LOAD:           /* Menu Item 'Load Map...' */
          if (load() == 0) redraw(NULL);
          return 0;             /* load a map and redraw the window */

        case MI_SAVE:           /* Menu Item 'Save Map...' */
          save(); return 0;     /* save a self-organizing map */

        case MI_QUIT:           /* Menu Item 'Quit' */
          DestroyWindow(hwnd);  /* stop network training */
          return 0;             /* and destroy the window */

        case MI_INIT:           /* Menu Item 'Initialize' */
          init(); redraw(NULL); /* initialize the map */
          return 0;             /* and redraw the window */

        case MI_START:          /* Menu Item 'Start Training' */
          mn_enable(0);         /* disable certain menu items */
          if (!som) init();     /* if there is no map, initialize */
          if ( som) { running = 1; train(hmain, WM_TIMER, 1, 0); }
          return 0;             /* start the training */

        case MI_STOP:           /* Menu Item 'Stop Training' */
          KillTimer(hmain, 1);  /* kill the timer */
          mn_enable(1);         /* enable all menu items */
          return running = 0;   /* clear the running flag */

        case MI_VIEW:           /* Menu Item 'Reset View' */
          p3d_eye  (proj3d, 0, 0, 5); /* init. eye, view, frustrum */
          p3d_view (proj3d, 0, -M_PI/2, 0);
          p3d_frust(proj3d, M_PI/6, 1e-6, 1e6);
          InvalidateRect(hwnd, NULL, 0);
          return 0;             /* invalidate window contents */

        case MI_REDRAW:         /* Menu Item 'Redraw' */
          InvalidateRect(hwnd, NULL, 0);
          return 0;             /* invalidate window contents */

        case MI_GRID:           /* Menu Item 'Grid/Shape...' */
          if (DialogBox(hinst, MKIRSC(DB_GRID), hwnd,
             (DLGPROC)dp_grid) == 0)
            InvalidateRect(hwnd, NULL, 0);
          return 0;             /* get pattern region and grid size */

        case MI_PARAMS:         /* Menu Item 'Parameters...' */
          if (DialogBox(hinst, MKIRSC(DB_PARAMS), hwnd,
             (DLGPROC)dp_params) == 0)
            InvalidateRect(hwnd, NULL, 0);
          return 0;             /* get learning parameters */

        case MI_ABOUT:          /* Menu Item 'About' */
          DialogBox(hinst, MKIRSC(DB_ABOUT), hwnd, (DLGPROC)dp_about);
          return 0;             /* show dialog box 'About Wsom' */
      } break;

    case WM_LBUTTONDOWN:        /* left   mouse button pressed */
      mode = (wp & MK_SHIFT) ? PITCH : ROTATE;
    case WM_MBUTTONDOWN:        /* middle mouse button pressed */
      if (msg == WM_MBUTTONDOWN) mode = ROLL;
    case WM_RBUTTONDOWN:        /* right  mouse button pressed */
      if (msg == WM_RBUTTONDOWN) mode = (wp & MK_SHIFT) ? YAW   : ZOOM;
      mx = LOWORD(lp); my = HIWORD(lp);
      p3d_save(proj3d);         /* save the current mouse position */
      return 0;                 /* and the current projection */

    case WM_LBUTTONUP:          /* mouse button released */
    case WM_MBUTTONUP:
    case WM_RBUTTONUP:
      mode = 0; return 0;       /* clear the movement mode */

    case WM_MOUSEMOVE:          /* if the mouse was moved */
      if (mode <= 0) break;
      mouse(LOWORD(lp), HIWORD(lp));
      InvalidateRect(hwnd, NULL, 0);
      return 0;                /* set new projection and redraw */

    case WM_CLOSE:              /* window closed */
    case WM_DESTROY:            /* or destroyed */
      PostQuitMessage(0);       /* terminate program */
      return 0;
  }
  return DefWindowProc(hwnd, msg, wp, lp);
}  /* wp_main() */

/*----------------------------------------------------------------------
  Main Program
----------------------------------------------------------------------*/

int WINAPI WinMain (HINSTANCE hcurr, HINSTANCE hprev,
                    LPSTR cmdln, int cmd)
{                               /* --- main function */
  WNDCLASS wc;                  /* window class */
  RECT     rc;                  /* window rectangle */
  MSG      msg;                 /* message */

  /* --- create window class --- */
  hinst = hcurr;                /* note instance handle */
  if (!hprev) {                 /* if first instance of this program */
    wc.style         = CS_HREDRAW | CS_VREDRAW;
    wc.lpfnWndProc   = wp_main;
    wc.cbClsExtra    = wc.cbWndExtra = 0;
    wc.hInstance     = hinst;
    wc.hIcon         = LoadIcon(NULL, IDI_APPLICATION);
    wc.hCursor       = LoadCursor(NULL, IDC_ARROW);
    wc.hbrBackground = NULL;
    wc.lpszMenuName  = MKIRSC(MN_MAIN);
    wc.lpszClassName = "WsomClass";
    if (!RegisterClass(&wc))    /* register class */
      return 0;                 /* of main window */
  }

  srand((unsigned long)time(NULL));
  proj3d = p3d_create();        /* create a projection object */
  if (!proj3d) error(E_NOMEM);  /* for sphere visualization */
  p3d_eye  (proj3d, 0, 0, 5);   /* initialize eye, view, and frustrum */
  p3d_view (proj3d, 0, -M_PI/2, 0);
  p3d_frust(proj3d, M_PI/6, 1e-6, 1e6);

  /* --- get pens and brushes --- */
  pen_black = GetStockObject(BLACK_PEN);
  pen_grey  = CreatePen(PS_SOLID, 1, RGB(0xc0, 0xc0, 0xc0));
  br_white  = GetStockObject(WHITE_BRUSH);

  /* --- create window --- */
  rc.left = 0; rc.right  = INITXEXT -1;
  rc.top  = 0; rc.bottom = INITYEXT -1;
  AdjustWindowRect(&rc, WINSTYLE, 1);
  hmain = CreateWindow("WsomClass", PRGNAME, WINSTYLE,
                       30, 30, rc.right-rc.left, rc.bottom-rc.top,
                       NULL, NULL, hinst, NULL);
  if (!hmain) return 0;         /* create main window */
  hdc_win = GetDC(hmain);
  if (!hdc_win) error(E_INIT);
  hdc_buf = CreateCompatibleDC(hdc_win);
  if (!hdc_buf) error(E_INIT);
  ShowWindow(hmain, cmd);       /* open and */
  UpdateWindow(hmain);          /* update main window */

  /* --- dispatch messages --- */
  while (GetMessage(&msg, NULL, 0, 0) > 0) {
    TranslateMessage(&msg);     /* get messages, translate */
    DispatchMessage (&msg);     /* and dispatch them */
  }

  /* --- clean up --- */
  ReleaseDC(hmain, hdc_win);    /* release device context of window */
  DeleteObject(hdc_buf);        /* delete device context of buffer */
  DeleteObject(buffer);         /* delete drawing buffer */
  DeleteObject(pen_grey);       /* delete created pen */
  return (int)msg.wParam;
}  /* WinMain() */
